package com.sun.xml.internal.bind.v2.runtime.property;

import com.sun.xml.internal.bind.v2.runtime.reflect.NullSafeAccessor;
import com.sun.xml.internal.bind.v2.runtime.reflect.Accessor;
import com.sun.xml.internal.bind.v2.runtime.Transducer;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.TextLoader;
import com.sun.xml.internal.bind.v2.model.core.PropertyKind;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.Loader;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.Receiver;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.DefaultValueLoaderDecorator;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.XsiNilLoader;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.ChildLoader;
import com.sun.xml.internal.bind.v2.util.QNameMap;
import com.sun.xml.internal.bind.api.AccessorException;
import org.xml.sax.SAXException;
import javax.xml.stream.XMLStreamException;
import java.io.IOException;
import com.sun.xml.internal.bind.v2.runtime.reflect.ListIterator;
import javax.xml.bind.JAXBException;
import com.sun.xml.internal.bind.v2.runtime.reflect.Lister;
import com.sun.xml.internal.bind.v2.runtime.XMLSerializer;
import java.util.Iterator;
import java.util.List;
import com.sun.xml.internal.bind.v2.runtime.RuntimeUtil;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeTypeRef;
import java.util.HashMap;
import com.sun.xml.internal.bind.v2.runtime.JAXBContextImpl;
import com.sun.xml.internal.bind.v2.runtime.Name;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeElementPropertyInfo;
import com.sun.xml.internal.bind.v2.runtime.JaxBeanInfo;
import java.lang.reflect.Type;
import com.sun.xml.internal.bind.v2.model.core.TypeRef;
import java.util.Map;

abstract class ArrayElementProperty<BeanT, ListT, ItemT> extends ArrayERProperty<BeanT, ListT, ItemT>
{
    private final Map<Class, TagAndType> typeMap;
    private Map<TypeRef<Type, Class>, JaxBeanInfo> refs;
    protected RuntimeElementPropertyInfo prop;
    private final Name nillableTagName;
    
    protected ArrayElementProperty(JAXBContextImpl grammar, RuntimeElementPropertyInfo prop) {
        super(grammar, prop, prop.getXmlName(), prop.isCollectionNillable());
        this.typeMap = new HashMap();
        this.refs = new HashMap();
        this.prop = prop;
        List<? extends RuntimeTypeRef> types = prop.getTypes();
        Name n = null;
        for (RuntimeTypeRef typeRef : types) {
            Class<Object> type = (Class)typeRef.getTarget().getType();
            if (type.isPrimitive()) {
                type = (Class)RuntimeUtil.primitiveToBox.get(type);
            }
            JaxBeanInfo beanInfo = grammar.getOrCreate(typeRef.getTarget());
            TagAndType tt = new TagAndType(grammar.nameBuilder.createElementName(typeRef.getTagName()), beanInfo);
            this.typeMap.put(type, tt);
            this.refs.put(typeRef, beanInfo);
            if (typeRef.isNillable() && n == null) {
                n = tt.tagName;
            }
            else {
                continue;
            }
        }
        this.nillableTagName = n;
    }
    
    @Override
    public void wrapUp() {
        super.wrapUp();
        this.refs = null;
        this.prop = null;
    }
    
    @Override
    protected void serializeListBody(BeanT beanT, XMLSerializer w, ListT list) throws IOException, XMLStreamException, SAXException, AccessorException {
        ListIterator<ItemT> itr = this.lister.iterator(list, w);
        boolean isIdref = itr instanceof Lister.IDREFSIterator;
        while (itr.hasNext()) {
            try {
                ItemT item = itr.next();
                if (item != null) {
                    Class<Object> itemType = item.getClass();
                    if (isIdref) {
                        itemType = ((Lister.IDREFSIterator)itr).last().getClass();
                    }
                    TagAndType tt;
                    for (tt = (TagAndType)this.typeMap.get(itemType); tt == null && itemType != null; itemType = itemType.getSuperclass(), tt = (TagAndType)this.typeMap.get(itemType)) {}
                    if (tt == null) {
                        w.startElement(((TagAndType)this.typeMap.values().iterator().next()).tagName, null);
                        w.childAsXsiType(item, this.fieldName, w.grammar.getBeanInfo(Object.class), false);
                    }
                    else {
                        w.startElement(tt.tagName, null);
                        this.serializeItem(tt.beanInfo, item, w);
                    }
                    w.endElement();
                }
                else if (this.nillableTagName != null) {
                    w.startElement(this.nillableTagName, null);
                    w.writeXsiNilTrue();
                    w.endElement();
                }
                else {
                    continue;
                }
            }
            catch (JAXBException e) {
                w.reportError(this.fieldName, e);
            }
        }
    }
    
    protected abstract void serializeItem(JaxBeanInfo p0, ItemT p1, XMLSerializer p2) throws SAXException, AccessorException, IOException, XMLStreamException;
    
    public void createBodyUnmarshaller(UnmarshallerChain chain, QNameMap<ChildLoader> loaders) {
        int offset = chain.allocateOffset();
        Receiver recv = new ArrayERProperty.ReceiverImpl(offset);
        for (RuntimeTypeRef typeRef : this.prop.getTypes()) {
            Name tagName = chain.context.nameBuilder.createElementName(typeRef.getTagName());
            Loader item = this.createItemUnmarshaller(chain, typeRef);
            if (typeRef.isNillable() || chain.context.allNillable) {
                item = new XsiNilLoader.Array(item);
            }
            if (typeRef.getDefaultValue() != null) {
                item = new DefaultValueLoaderDecorator(item, typeRef.getDefaultValue());
            }
            loaders.put(tagName, new ChildLoader(item, recv));
        }
    }
    
    @Override
    public final PropertyKind getKind() {
        return PropertyKind.ELEMENT;
    }
    
    private Loader createItemUnmarshaller(UnmarshallerChain chain, RuntimeTypeRef typeRef) {
        if (PropertyFactory.isLeaf(typeRef.getSource())) {
            Transducer xducer = typeRef.getTransducer();
            return new TextLoader(xducer);
        }
        return ((JaxBeanInfo)this.refs.get(typeRef)).getLoader(chain.context, true);
    }
    
    @Override
    public Accessor getElementPropertyAccessor(String nsUri, String localName) {
        if (this.wrapperTagName != null) {
            if (this.wrapperTagName.equals(nsUri, localName)) {
                return this.acc;
            }
        }
        else {
            for (TagAndType tt : this.typeMap.values()) {
                if (tt.tagName.equals(nsUri, localName)) {
                    return new NullSafeAccessor(this.acc, this.lister);
                }
                continue;
            }
        }
        return null;
    }
}
